#include <fcntl.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>
#include <time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <CL/opencl.h>

void nextkey();

#define NUM_BLOCKS 102400
#define BLOCK_SIZE 256
#define CHARSET "abcdefghijklmnopqrstuvwxyz1234567890"

#ifndef uint32_t
#define uint32_t unsigned int
#endif

typedef struct {
    uint32_t h0,h1,h2,h3,h4;
} SHA_DEV_CTX;

char keybuf[64+1];

int main(int argc, char **argv) {
   char c;
   unsigned int i=0, j, x;
   char *textstring = (char *)malloc(64*BLOCK_SIZE);
   char *result = (char *)malloc(64*BLOCK_SIZE);
   static SHA_DEV_CTX *outbuffer;
   clock_t start;
   size_t max_group_size;



   memset(&keybuf,0,sizeof(keybuf));
   strncpy(keybuf,CHARSET,1);

   if (textstring == NULL || result == NULL) {
      printf("Couldn't allocate memory.\n");
      return 0;
   }

   int err;
   unsigned int length[BLOCK_SIZE];

   size_t global = BLOCK_SIZE;
   size_t local = BLOCK_SIZE;

   cl_platform_id platform;
   cl_device_id device_id;
   cl_context context;
   cl_command_queue commands;
   cl_program program;
   cl_kernel kernel;

   cl_mem input;
   cl_mem devlen;
   cl_mem output;
   cl_mem pinned_partial_hashes;

   int gpu = 1;
   if(argc == 2){
	printf("Using CPU\n");
	gpu = 0;
   }
   // load kernel source
   char *source=(char*)malloc(1024*16);

   FILE *fp = fopen("./ssha_opencl_kernel.cl","r");
   if(!fp){
	   printf("Source kernel not found!\n");
	   exit(2);
   }
   size_t source_size = fread(source, sizeof(char), 1024*16, fp);
   source[source_size] = 0;
   fclose(fp);

   char log[1024*64];
   err = clGetPlatformIDs(1, &platform, NULL);
   err = clGetPlatformInfo(platform, CL_PLATFORM_NAME, sizeof(log), log, NULL);
   printf("\nOpenCL Platform: <<<%s>>>\n", log);
   err = clGetDeviceIDs(platform, gpu ? CL_DEVICE_TYPE_GPU : CL_DEVICE_TYPE_CPU, 1, &device_id, NULL);
   if (err != CL_SUCCESS) {
            printf("Error: Failed to create a device group!\n");
            return EXIT_FAILURE;
   }
   err = clGetDeviceInfo(device_id, CL_DEVICE_NAME, sizeof(log), log, NULL);
   printf(" and device: <<<%s>>>\n",log);

   err = clGetDeviceInfo(device_id, CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(max_group_size), &max_group_size, NULL);
   printf("max group size %d\n",max_group_size); 



   context = clCreateContext(0, 1, &device_id, NULL, NULL, &err);
   if (!context) {
            printf("Error: Failed to create a compute context!\n");
            return EXIT_FAILURE;
   }

   commands = clCreateCommandQueue(context, device_id, 0, &err);
   if (!commands) {
            printf("Error: Failed to create a command commands!\n");
            return EXIT_FAILURE;
   }

   program = clCreateProgramWithSource(context, 1, (const char **) &source, NULL, &err);
   if (!program) {
            printf("Error: Failed to create compute program! (error %d)\n",err);
            switch (err) {
               case CL_INVALID_CONTEXT: printf("context is not a valid context.\n"); break;
               case CL_INVALID_VALUE: printf("count is zero or if strings or any entry in strings is NULL\n"); break;
               case CL_OUT_OF_HOST_MEMORY: printf("there is a failure to allocate resources required by the OpenCL implementation on the host\n"); break;
            }
            return EXIT_FAILURE;
      }

  err = clBuildProgram(program, 1, &device_id, NULL, NULL, NULL);
      if (err != CL_SUCCESS)
      {
            size_t len;
            char buffer[2048];

            printf("Error: Failed to build program executable!\n");
            clGetProgramBuildInfo(program, device_id, CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, &len);
            printf("%s\n", buffer);
            exit(1);
      }

      kernel = clCreateKernel(program, "sha1_crypt_kernel", &err);
      if (!kernel || err != CL_SUCCESS)
      {
            printf("Error: Failed to create compute kernel!\n");
            switch (err) {
               case CL_INVALID_PROGRAM: printf("if program is not a valid program object\n"); break;
               case CL_INVALID_PROGRAM_EXECUTABLE: printf("if there is no successfully built executable for program\n"); break;
               case CL_INVALID_KERNEL_NAME: printf("if kernel_name is not found in program\n"); break;
               case CL_INVALID_KERNEL_DEFINITION: printf("if the function definition for __kernel function given by kernel_name such as the number of arguments, the argument types are not the same for all devices for which the program executable has been built\n"); break;
               case CL_INVALID_VALUE: printf("is kernel_name is NULL\n"); break;
               case CL_OUT_OF_HOST_MEMORY: printf("if there is a failure to allocate resources required by the OpenCL implementation on the host\n"); break;
            }
            exit(1);
      }

      pinned_partial_hashes = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, sizeof(SHA_DEV_CTX)*BLOCK_SIZE, NULL, &err);
      if(err != CL_SUCCESS){
	printf("pinned ahio %d\n",err);
      }

      outbuffer = (SHA_DEV_CTX*)clEnqueueMapBuffer(commands, pinned_partial_hashes, CL_TRUE, CL_MAP_READ, 0, sizeof(SHA_DEV_CTX)*BLOCK_SIZE, 0, NULL, NULL, &err);
      if(err != CL_SUCCESS){
	printf("outbuffer ahio %d\n",err);
      }


      input = clCreateBuffer(context,  CL_MEM_READ_ONLY,   64*BLOCK_SIZE, NULL, NULL);
      output = clCreateBuffer(context, CL_MEM_WRITE_ONLY, sizeof(SHA_DEV_CTX)*BLOCK_SIZE, NULL, NULL);
      devlen = clCreateBuffer(context, CL_MEM_READ_ONLY, sizeof(unsigned int)*BLOCK_SIZE, NULL, NULL);

      if (!input || !output)
      {
            printf("Error: Failed to allocate device memory!\n");
            exit(1);
      }

      err = 0;
      err  = clSetKernelArg(kernel, 0, sizeof(cl_mem), &input);
      err |= clSetKernelArg(kernel, 1, sizeof(cl_mem), &devlen);
      err |= clSetKernelArg(kernel, 2, sizeof(cl_mem), &output);
      if (err != CL_SUCCESS)
      {
            printf("Error: Failed to set kernel arguments! %d\n", err);
            switch (err) {
               case CL_INVALID_KERNEL: printf("kernel is not a valid kernel object\n"); break;
               case CL_INVALID_ARG_INDEX: printf("arg_index is not a valid argument index\n"); break;
               case CL_INVALID_ARG_VALUE: printf("arg_value specified is NULL for an argument that is not declared with the __local qualifier or vice-versa\n"); break;
               case CL_INVALID_MEM_OBJECT: printf("an argument declared to be a memory object when the specified arg_value is not a valid memory object\n"); break;
               case CL_INVALID_SAMPLER: printf("an argument declared to be of type sampler_t when the specified arg_value is not a valid sampler object\n"); break;
               case CL_INVALID_ARG_SIZE: printf("arg_size does not match the size of the data type for an argument that is not a memory object or if the argument is a memory object and arg_size != sizeof(cl_mem) or if arg_size is zero and the argument is declared with the __local qualifier or if the argument is a sampler and arg_size != sizeof(cl_sampler)\n"); break;
            }
            exit(1);
      }
      start = clock();
      for (x = 0; x<NUM_BLOCKS; x++) {
/*	if(( x % (NUM_BLOCKS/10)) == 0){
		printf("we are at %f %%\n",((double) x*BLOCK_SIZE)/( (double) NUM_BLOCKS*BLOCK_SIZE));
	} */
         for (i=0; i<BLOCK_SIZE; i++) {
            strcpy(&textstring[i*64],keybuf);
            length[i] = strlen(&textstring[i*64]);
            nextkey();
         }

         err = clEnqueueWriteBuffer(commands, input, CL_TRUE, 0, 64*BLOCK_SIZE, textstring, 0, NULL, NULL);
         if (err != CL_SUCCESS)
         {
               printf("Error: Failed to write to source array!\n");
               exit(1);
         }

         err = clEnqueueWriteBuffer(commands, devlen, CL_TRUE, 0, sizeof(unsigned int)*BLOCK_SIZE, length, 0, NULL, NULL);
         if (err != CL_SUCCESS)
         {
               printf("Error: Failed to write to source array!\n");
               exit(1);
         }

         err = clEnqueueNDRangeKernel(commands, kernel, 1, NULL, &global, &local, 0, NULL, NULL);
         if (err)
         {
               printf("Error: Failed to execute kernel!\n");
               return EXIT_FAILURE;
         }

         clFinish(commands);

         err = clEnqueueReadBuffer( commands, output, CL_TRUE, 0, sizeof(SHA_DEV_CTX)*BLOCK_SIZE, outbuffer, 0, NULL, NULL );
         if (err != CL_SUCCESS){
               printf("Error: Failed to read output array! %d\n", err);
               exit(1);
         }
	 (*outbuffer).h0 = be32toh((*outbuffer).h0);
	 (*outbuffer).h1 = be32toh((*outbuffer).h1);
	 (*outbuffer).h2 = be32toh((*outbuffer).h2);
	 (*outbuffer).h3 = be32toh((*outbuffer).h3);
	 (*outbuffer).h4 = be32toh((*outbuffer).h4);
         //char *result = (char *)malloc(64*256);
	 
      }

      //printf("Done, last result was: %s - %s\n",&textstring[(i-1)*64],&result[(i-1)*64]);
      printf("Computed %f hashes.\n",((double) NUM_BLOCKS*BLOCK_SIZE));
      printf ( "%f\n", ( (double)clock() - start ) / CLOCKS_PER_SEC );


      clReleaseMemObject(input);
      clReleaseMemObject(output);
      clReleaseProgram(program);
      clReleaseKernel(kernel);
      clReleaseCommandQueue(commands);
      clReleaseContext(context);
   
      return 0;
   }

   void nextkey() {
      int i,j,k,found;

      if (keybuf[strlen(keybuf)-1] == CHARSET[strlen(CHARSET)-1])
      {
         found=0;
         for (i=strlen(keybuf)-1;i>=0;i--)
         {
            if (keybuf[i] != CHARSET[strlen(CHARSET)-1])
            {
               for (j=0;j<strlen(CHARSET);j++)
               {
                  if (keybuf[i] == CHARSET[j])
                  {
                     keybuf[i] = CHARSET[j+1];
                     for (k=i+1;k<strlen(keybuf);k++)
                     {
                        keybuf[k] = CHARSET[0];
                     }
                     found=1;
                     break;
                  }
               }
               i=-1;
            }
         }
         if (!found)
         {
            for (i=0;i<strlen(keybuf);i++)
            {
               keybuf[i] = CHARSET[0];
            }
            strncat(keybuf,CHARSET,1);
         }
      }
      else
      {
         for (i=0;i<strlen(CHARSET);i++)
         {
            if (keybuf[strlen(keybuf)-1] == CHARSET[i]) break;
         }
         keybuf[strlen(keybuf)-1] = CHARSET[i+1];
      }
   }
